@@ -79,22 +79,30 @@ func isTypeLit(t Type) bool {
 }
 
 // isTyped reports whether t is typed; i.e., not an untyped
-// constant or boolean. isTyped may be called with types that
-// are not fully set up.
+// constant or boolean.
+// Safe to call from types that are not fully set up.
 func isTyped(t Type) bool {
-	// Alias or Named types cannot denote untyped types,
-	// thus we don't need to call Unalias or under
-	// (which would be unsafe to do for types that are
-	// not fully set up).
+	// Alias and named types cannot denote untyped types
+	// so there's no need to call Unalias or under, below.
 	b, _ := t.(*Basic)
 	return b == nil || b.info&IsUntyped == 0
 }
 
 // isUntyped(t) is the same as !isTyped(t).
+// Safe to call from types that are not fully set up.
 func isUntyped(t Type) bool {
 	return !isTyped(t)
 }
 
+// isUntypedNumeric reports whether t is an untyped numeric type.
+// Safe to call from types that are not fully set up.
+func isUntypedNumeric(t Type) bool {
+	// Alias and named types cannot denote untyped types
+	// so there's no need to call Unalias or under, below.
+	b, _ := t.(*Basic)
+	return b != nil && b.info&IsUntyped != 0 && b.info&IsNumeric != 0
+}
+
 // IsInterface reports whether t is an interface type.
 func IsInterface(t Type) bool {
 	_, ok := under(t).(*Interface)
@@ -129,6 +137,9 @@ func hasEmptyTypeset(t Type) bool {
 // TODO(gri) should we include signatures or assert that they are not present?
 func isGeneric(t Type) bool {
 	// A parameterized type is only generic if it doesn't have an instantiation already.
+	if alias, _ := t.(*Alias); alias != nil && alias.tparams != nil && alias.targs == nil {
+		return true
+	}
 	named := asNamed(t)
 	return named != nil && named.obj != nil && named.inst == nil && named.TypeParams().Len() > 0
 }
@@ -205,6 +216,16 @@ func hasNil(t Type) bool {
 	return false
 }
 
+// samePkg reports whether packages a and b are the same.
+func samePkg(a, b *Package) bool {
+	// package is nil for objects in universe scope
+	if a == nil || b == nil {
+		return a == b
+	}
+	// a != nil && b != nil
+	return a.path == b.path
+}
+
 // An ifacePair is a node in a stack of interface type pairs compared for identity.
 type ifacePair struct {
 	x, y *Interface
@@ -269,7 +290,7 @@ func (c *comparer) identical(x, y Type, p *ifacePair) bool {
 					g := y.fields[i]
 					if f.embedded != g.embedded ||
 						!c.ignoreTags && x.Tag(i) != y.Tag(i) ||
-						!f.sameId(g.pkg, g.name) ||
+						!f.sameId(g.pkg, g.name, false) ||
 						!c.identical(f.typ, g.typ, p) {
 						return false
 					}
@@ -467,7 +488,7 @@ func (c *comparer) identical(x, y Type, p *ifacePair) bool {
 		// avoid a crash in case of nil type
 
 	default:
-		unreachable()
+		panic("unreachable")
 	}
 
 	return false
@@ -500,7 +521,9 @@ func identicalInstance(xorig Type, xargs []Type, yorig Type, yargs []Type) bool
 // it returns the incoming type for all other types. The default type
 // for untyped nil is untyped nil.
 func Default(t Type) Type {
-	if t, ok := Unalias(t).(*Basic); ok {
+	// Alias and named types cannot denote untyped types
+	// so there's no need to call Unalias or under, below.
+	if t, _ := t.(*Basic); t != nil {
 		switch t.kind {
 		case UntypedBool:
 			return Typ[Bool]
@@ -529,7 +552,7 @@ func maxType(x, y Type) Type {
 	if x == y {
 		return x
 	}
-	if isUntyped(x) && isUntyped(y) && isNumeric(x) && isNumeric(y) {
+	if isUntypedNumeric(x) && isUntypedNumeric(y) {
 		// untyped types are basic types
 		if x.(*Basic).kind > y.(*Basic).kind {
 			return x
