@@ -1,5 +1,5 @@
 // cmd/7l/noop.c, cmd/7l/obj.c, cmd/ld/pass.c from Vita Nuova.
-// https://code.google.com/p/ken-cc/source/browse/
+// https://bitbucket.org/plan9-from-bell-labs/9-cc/src/master/
 //
 // 	Copyright © 1994-1999 Lucent Technologies Inc. All rights reserved.
 // 	Portions Copyright © 1995-1997 C H Forsyth (forsyth@terzarima.net)
@@ -65,6 +65,9 @@ var zrReplace = map[obj.As]bool{
 }
 
 func (c *ctxt7) stacksplit(p *obj.Prog, framesize int32) *obj.Prog {
+	// Jump back to here after morestack returns.
+	startPred := c.cursym.Func().Text
+
 	if c.ctxt.Flag_maymorestack != "" {
 		p = c.cursym.Func().SpillRegisterArgs(p, c.newprog)
 
@@ -145,11 +148,9 @@ func (c *ctxt7) stacksplit(p *obj.Prog, framesize int32) *obj.Prog {
 		p.Spadj = -frameSize
 
 		p = c.cursym.Func().UnspillRegisterArgs(p, c.newprog)
+		startPred = p
 	}
 
-	// Jump back to here after morestack returns.
-	startPred := p
-
 	// MOV	g_stackguard(g), RT1
 	p = obj.Appendp(p, c.newprog)
 
@@ -289,11 +290,12 @@ func (c *ctxt7) stacksplit(p *obj.Prog, framesize int32) *obj.Prog {
 	}
 	call.To.Sym = c.ctxt.Lookup(morestack)
 
-	unspill := c.cursym.Func().UnspillRegisterArgs(call, c.newprog)
-	pcdata = c.ctxt.EndUnsafePoint(unspill, c.newprog, -1)
+	// The instructions which unspill regs should be preemptible.
+	pcdata = c.ctxt.EndUnsafePoint(call, c.newprog, -1)
+	unspill := c.cursym.Func().UnspillRegisterArgs(pcdata, c.newprog)
 
 	// B	start
-	jmp := obj.Appendp(pcdata, c.newprog)
+	jmp := obj.Appendp(unspill, c.newprog)
 	jmp.As = AB
 	jmp.To.Type = obj.TYPE_BRANCH
 	jmp.To.SetTarget(startPred.Link)
@@ -329,8 +331,33 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 		break
 	}
 
-	// Rewrite float constants to values stored in memory.
+	// Rewrite float and vector constants to values stored in memory.
 	switch p.As {
+	case AVMOVS:
+		if p.From.Type == obj.TYPE_CONST {
+			p.From.Type = obj.TYPE_MEM
+			p.From.Sym = c.ctxt.Int32Sym(p.From.Offset)
+			p.From.Name = obj.NAME_EXTERN
+			p.From.Offset = 0
+		}
+
+	case AVMOVD:
+		if p.From.Type == obj.TYPE_CONST {
+			p.From.Type = obj.TYPE_MEM
+			p.From.Sym = c.ctxt.Int64Sym(p.From.Offset)
+			p.From.Name = obj.NAME_EXTERN
+			p.From.Offset = 0
+		}
+
+	case AVMOVQ:
+		if p.From.Type == obj.TYPE_CONST {
+			p.From.Type = obj.TYPE_MEM
+			p.From.Sym = c.ctxt.Int128Sym(p.GetFrom3().Offset, p.From.Offset)
+			p.From.Name = obj.NAME_EXTERN
+			p.From.Offset = 0
+			p.RestArgs = nil
+		}
+
 	case AFMOVS:
 		if p.From.Type == obj.TYPE_FCONST {
 			f64 := p.From.Val.(float64)
@@ -365,8 +392,6 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 			p.From.Name = obj.NAME_EXTERN
 			p.From.Offset = 0
 		}
-
-		break
 	}
 
 	if c.ctxt.Flag_dynlink {
@@ -485,6 +510,24 @@ func (c *ctxt7) rewriteToUseGot(p *obj.Prog) {
 	obj.Nopout(p)
 }
 
+var runtimeEntryPoints = map[string]struct{}{
+	"runtime.(*_panic).start": {},
+	"runtime.deferproc": {},
+	"runtime.deferprocStack": {},
+	"runtime.mcall": {},
+	"runtime.morestack": {},
+	"runtime.morestack_noctxt": {},
+	"runtime.mstart1": {},
+}
+
+func isRuntimeEntryPoint(s *obj.LSym) bool {
+	if s == nil {
+		return false
+	}
+	_, ok := runtimeEntryPoints[s.Name]
+	return ok
+}
+
 func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 	if cursym.Func().Text == nil || cursym.Func().Text.Link == nil {
 		return
@@ -528,20 +571,20 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 
 	var q *obj.Prog
 	var q1 *obj.Prog
-	var retjmp *obj.LSym
+	var last *obj.Prog
 	for p := c.cursym.Func().Text; p != nil; p = p.Link {
-		o := p.As
-		switch o {
+		switch p.As {
 		case obj.ATEXT:
-			c.cursym.Func().Text = p
+			text := p
+			cursym.Func().Text = text
 			c.autosize = int32(textstksiz)
 
-			if p.Mark&LEAF != 0 && c.autosize == 0 {
+			if text.Mark&LEAF != 0 && c.autosize == 0 {
 				// A leaf function with no locals has no frame.
-				p.From.Sym.Set(obj.AttrNoFrame, true)
+				text.From.Sym.Set(obj.AttrNoFrame, true)
 			}
 
-			if !p.From.Sym.NoFrame() {
+			if !text.From.Sym.NoFrame() {
 				// If there is a stack frame at all, it includes
 				// space to save the LR.
 				c.autosize += 8
@@ -563,10 +606,10 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 
 				// low 32 bits for autosize
 				// high 32 bits for extrasize
-				p.To.Offset = int64(c.autosize) | int64(extrasize)<<32
+				text.To.Offset = int64(c.autosize) | int64(extrasize)<<32
 			} else {
 				// NOFRAME
-				p.To.Offset = 0
+				text.To.Offset = 0
 			}
 
 			if c.autosize == 0 && c.cursym.Func().Text.Mark&LEAF == 0 {
@@ -576,20 +619,28 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 				c.cursym.Func().Text.Mark |= LEAF
 			}
 
+			// Add branch target indicator for function entry.
+			// TODO(jsing): May be able to use SPOP_C outside of runtime (which
+			// makes use of systemstack).
+			p = obj.Appendp(text, newprog)
+			p.As = ABTI
+			p.From.Type = obj.TYPE_SPECIAL
+			p.From.Offset = int64(SPOP_JC)
+
 			if cursym.Func().Text.Mark&LEAF != 0 {
 				cursym.Set(obj.AttrLeaf, true)
-				if p.From.Sym.NoFrame() {
+				if text.From.Sym.NoFrame() {
 					break
 				}
 			}
 
-			if p.Mark&LEAF != 0 && c.autosize < abi.StackSmall {
+			if text.Mark&LEAF != 0 && c.autosize < abi.StackSmall {
 				// A leaf function with a small stack can be marked
 				// NOSPLIT, avoiding a stack check.
-				p.From.Sym.Set(obj.AttrNoSplit, true)
+				text.From.Sym.Set(obj.AttrNoSplit, true)
 			}
 
-			if !p.From.Sym.NoSplit() {
+			if !text.From.Sym.NoSplit() {
 				p = c.stacksplit(p, c.autosize) // emit split check
 			}
 
@@ -822,25 +873,31 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 				break
 			}
 
-			retjmp = p.To.Sym
+			retJMP, retReg := p.To.Sym, p.To.Reg
+			if retReg == 0 {
+				retReg = REGLINK
+			}
 			p.To = obj.Addr{}
 			if c.cursym.Func().Text.Mark&LEAF != 0 {
 				if c.autosize != 0 {
+					// Restore frame pointer.
+					// ADD $framesize-8, RSP, R29
+					p.As = AADD
+					p.From.Type = obj.TYPE_CONST
+					p.From.Offset = int64(c.autosize) - 8
+					p.Reg = REGSP
+					p.To.Type = obj.TYPE_REG
+					p.To.Reg = REGFP
+
+					// Pop stack frame.
+					// ADD $framesize, RSP, RSP
+					p = obj.Appendp(p, c.newprog)
 					p.As = AADD
 					p.From.Type = obj.TYPE_CONST
 					p.From.Offset = int64(c.autosize)
 					p.To.Type = obj.TYPE_REG
 					p.To.Reg = REGSP
 					p.Spadj = -c.autosize
-
-					// Frame pointer.
-					p = obj.Appendp(p, c.newprog)
-					p.As = ASUB
-					p.From.Type = obj.TYPE_CONST
-					p.From.Offset = 8
-					p.Reg = REGSP
-					p.To.Type = obj.TYPE_REG
-					p.To.Reg = REGFP
 				}
 			} else {
 				aoffset := c.autosize
@@ -897,10 +954,10 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 				p = q
 			}
 
-			if retjmp != nil { // retjmp
+			if retJMP != nil {
 				p.As = AB
 				p.To.Type = obj.TYPE_BRANCH
-				p.To.Sym = retjmp
+				p.To.Sym = retJMP
 				p.Spadj = +c.autosize
 				break
 			}
@@ -908,7 +965,7 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 			p.As = obj.ARET
 			p.To.Type = obj.TYPE_MEM
 			p.To.Offset = 0
-			p.To.Reg = REGLINK
+			p.To.Reg = retReg
 			p.Spadj = +c.autosize
 
 		case AADD, ASUB:
@@ -1036,6 +1093,21 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 			p = q5
 		}
 
+		if p.As == obj.ACALL && isRuntimeEntryPoint(p.To.Sym) {
+			// Append BTI as we'll return after this call via gogo.
+			p = obj.Appendp(p, newprog)
+			p.As = ABTI
+			p.From.Type = obj.TYPE_SPECIAL
+			p.From.Offset = int64(SPOP_J)
+		}
+		if p.As == obj.ACALL && p.To.Sym != nil && p.To.Sym.Name == "runtime.deferreturn" && last != nil {
+			// Prepend BTI as we'll enter this point via gogo.
+			pp := obj.Appendp(last, newprog)
+			pp.As = ABTI
+			pp.From.Type = obj.TYPE_SPECIAL
+			pp.From.Offset = int64(SPOP_J)
+		}
+
 		if p.To.Type == obj.TYPE_REG && p.To.Reg == REGSP && p.Spadj == 0 {
 			f := c.cursym.Func()
 			if f.FuncFlag&abi.FuncFlagSPWrite == 0 {
@@ -1068,6 +1140,8 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 			p.From.Reg = int16(REG_LSL + r + (shift&7)<<5)
 			p.From.Offset = 0
 		}
+
+		last = p
 	}
 }
 
