@@ -320,29 +320,27 @@ const (
 
 	// Loong64.
 
-	// R_ADDRLOONG64 resolves to the low 12 bits of an external address, by encoding
-	// it into the instruction.
-	R_ADDRLOONG64
-
-	// R_ADDRLOONG64U resolves to the sign-adjusted "upper" 20 bits (bit 5-24) of an
+	// R_LOONG64_ADDR_HI resolves to the sign-adjusted "upper" 20 bits (bit 5-24) of an
 	// external address, by encoding it into the instruction.
-	R_ADDRLOONG64U
+	// R_LOONG64_ADDR_LO resolves to the low 12 bits of an external address, by encoding
+	// it into the instruction.
+	R_LOONG64_ADDR_HI
+	R_LOONG64_ADDR_LO
 
-	// R_ADDRLOONG64TLS resolves to the low 12 bits of a TLS address (offset from
+	// R_LOONG64_TLS_LE_HI resolves to the high 20 bits of a TLS address (offset from
 	// thread pointer), by encoding it into the instruction.
-	R_ADDRLOONG64TLS
-
-	// R_ADDRLOONG64TLSU resolves to the high 20 bits of a TLS address (offset from
+	// R_LOONG64_TLS_LE_LO resolves to the low 12 bits of a TLS address (offset from
 	// thread pointer), by encoding it into the instruction.
-	R_ADDRLOONG64TLSU
+	R_LOONG64_TLS_LE_HI
+	R_LOONG64_TLS_LE_LO
 
 	// R_CALLLOONG64 resolves to non-PC-relative target address of a CALL (BL/JIRL)
 	// instruction, by encoding the address into the instruction.
 	R_CALLLOONG64
 
-	// R_LOONG64_TLS_IE_PCREL_HI and R_LOONG64_TLS_IE_LO relocates a pcalau12i, ld.d
+	// R_LOONG64_TLS_IE_HI and R_LOONG64_TLS_IE_LO relocates a pcalau12i, ld.d
 	// pair to compute the address of the GOT slot of the tls symbol.
-	R_LOONG64_TLS_IE_PCREL_HI
+	R_LOONG64_TLS_IE_HI
 	R_LOONG64_TLS_IE_LO
 
 	// R_LOONG64_GOT_HI and R_LOONG64_GOT_LO resolves a GOT-relative instruction sequence,
@@ -357,25 +355,10 @@ const (
 	// R_ADDRMIPSU (only used on mips/mips64) resolves to the sign-adjusted "upper" 16
 	// bits (bit 16-31) of an external address, by encoding it into the instruction.
 	R_ADDRMIPSU
-
 	// R_ADDRMIPSTLS (only used on mips64) resolves to the low 16 bits of a TLS
 	// address (offset from thread pointer), by encoding it into the instruction.
 	R_ADDRMIPSTLS
 
-	// R_MIPS_GPREL_HI16 resolves to the upper 16 bits of a GP relative address.
-	R_MIPS_GPREL_HI16
-
-	// R_MIPS_GPREL_LO16 resolves to the lower 16 bits of a GP relative address.
-	R_MIPS_GPREL_LO16
-
-	// R_MIPS_CALL16 resolves to a 16 bit offset into the GOT, which contains the
-	// the address for a PLT entry.
-	R_MIPS_CALL16
-
-	// R_MIPS_JALR marks a JALR instruction of an external call, which may be
-	// optimized by the external linker.
-	R_MIPS_JALR
-
 	// R_ADDRCUOFF resolves to a pointer-sized offset from the start of the
 	// symbol's DWARF compile unit.
 	R_ADDRCUOFF
