@@ -398,6 +398,13 @@ func adddynrel(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 			// (e.g. go version).
 			return true
 		}
+	case objabi.R_GOTPCREL:
+		if target.IsExternal() {
+			// External linker will do this relocation.
+			return true
+		}
+		// We only need to handle external linking mode, as R_GOTPCREL can
+		// only occur in plugin or shared build modes.
 	}
 
 	return false
@@ -617,15 +624,12 @@ func addpltsym(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 			panic("plt is not set up")
 		}
 
-		// jmpq *got+size(IP)
-		plt.AddUint8(0xff)
-
-		plt.AddUint8(0x25)
-		plt.AddPCRelPlus(target.Arch, got.Sym(), got.Size())
-
 		// add to got: pointer to current pos in plt
 		got.AddAddrPlus(target.Arch, plt.Sym(), plt.Size())
 
+		// endbr64
+		plt.AddBytes([]byte{0xf3, 0x0f, 0x1e, 0xfa})
+
 		// pushq $x
 		plt.AddUint8(0x68)
 
@@ -636,6 +640,23 @@ func addpltsym(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 
 		plt.AddUint32(target.Arch, uint32(-(plt.Size() + 4)))
 
+		// nop (2 bytes)
+		plt.AddBytes([]byte{0x66, 0x90})
+
+		// .plt.sec equivalent
+
+		// endbr64
+		plt.AddBytes([]byte{0xf3, 0x0f, 0x1e, 0xfa})
+
+		// jmpq *got+size(IP)
+		plt.AddUint8(0xff)
+
+		plt.AddUint8(0x25)
+		plt.AddPCRelPlus(target.Arch, got.Sym(), got.Size()-8)
+
+		// nop (6 bytes)
+		plt.AddBytes([]byte{0x66, 0x0f, 0x1f, 0x44, 0x00, 0x00})
+
 		// rela
 		rela.AddAddrPlus(target.Arch, got.Sym(), got.Size()-8)
 
