@@ -6,7 +6,7 @@ package runtime
 
 import (
 	"internal/abi"
-	"runtime/internal/atomic"
+	"internal/runtime/atomic"
 	"unsafe"
 )
 
@@ -32,9 +32,6 @@ var sigset_all = ^sigset(0)
 
 // From OpenBSD's <sys/sysctl.h>
 const (
-	_CTL_KERN   = 1
-	_KERN_OSREV = 3
-
 	_CTL_HW        = 6
 	_HW_NCPU       = 3
 	_HW_PAGESIZE   = 7
@@ -86,13 +83,6 @@ func getPageSize() uintptr {
 	return 0
 }
 
-func getOSRev() int {
-	if osrev, ok := sysctlInt([]uint32{_CTL_KERN, _KERN_OSREV}); ok {
-		return int(osrev)
-	}
-	return 0
-}
-
 //go:nosplit
 func semacreate(mp *m) {
 }
@@ -147,17 +137,16 @@ func semawakeup(mp *m) {
 func osinit() {
 	ncpu = getncpu()
 	physPageSize = getPageSize()
-	haveMapStack = getOSRev() >= 201805 // OpenBSD 6.3
 }
 
 var urandom_dev = []byte("/dev/urandom\x00")
 
 //go:nosplit
-func getRandomData(r []byte) {
+func readRandom(r []byte) int {
 	fd := open(&urandom_dev[0], 0 /* O_RDONLY */, 0)
 	n := read(fd, unsafe.Pointer(&r[0]), int32(len(r)))
 	closefd(fd)
-	extendRandom(r, int(n))
+	return int(n)
 }
 
 func goenvs() {
@@ -187,6 +176,7 @@ func minit() {
 //go:nosplit
 func unminit() {
 	unminitSignals()
+	getg().m.procid = 0
 }
 
 // Called from exitm, but not from drop, to undo the effect of thread-owned
@@ -263,15 +253,7 @@ func validSIGPROF(mp *m, c *sigctxt) bool {
 	return true
 }
 
-var haveMapStack = false
-
 func osStackAlloc(s *mspan) {
-	// OpenBSD 6.4+ requires that stacks be mapped with MAP_STACK.
-	// It will check this on entry to system calls, traps, and
-	// when switching to the alternate system stack.
-	//
-	// This function is called before s is used for any data, so
-	// it's safe to simply re-map it.
 	osStackRemap(s, _MAP_STACK)
 }
 
@@ -281,13 +263,6 @@ func osStackFree(s *mspan) {
 }
 
 func osStackRemap(s *mspan, flags int32) {
-	if !haveMapStack {
-		// OpenBSD prior to 6.3 did not have MAP_STACK and so
-		// the following mmap will fail. But it also didn't
-		// require MAP_STACK (obviously), so there's no need
-		// to do the mmap.
-		return
-	}
 	a, err := mmap(unsafe.Pointer(s.base()), s.npages*pageSize, _PROT_READ|_PROT_WRITE, _MAP_PRIVATE|_MAP_ANON|_MAP_FIXED|flags, -1, 0)
 	if err != 0 || uintptr(a) != s.base() {
 		print("runtime: remapping stack memory ", hex(s.base()), " ", s.npages*pageSize, " a=", a, " err=", err, "\n")
